function [Vxq, Vyq, TTCq, Zxq, Zyq] = extractFlowFromQuadraticEquation(puq, pvq, WIDTH, HEIGHT)
% function [Vxq, Vyq, TTCq, Zxq, Zyq] = extractFlowFromQuadraticEquation(puq, pvq, WIDTH, HEIGHT)
% Extract ventral flow, time-to-contact and slopes from quadratic equation

threshold_flow_translational = 1.0;
threshold_rel_vert = 0.0; %0.04;

% slope can be estimated with horizontal movement component or vertical
% movement component:
Vxq = -(puq(1) + (WIDTH/2) * puq(2) + (HEIGHT/2) * puq(3) + (WIDTH/2) * (WIDTH/2) * puq(4) + (WIDTH/2) * (HEIGHT/2) * puq(5));
Vyq = -(pvq(1) + (WIDTH/2) * pvq(2) + (HEIGHT/2) * pvq(3) + (HEIGHT/2) * (HEIGHT/2) * pvq(4) + (WIDTH/2) * (HEIGHT/2) * pvq(5));

CRISP = true;

% u = pu(1) + pu(2) * x + pu(3) * y + pu(4) * x^2 + pu(5) * xy
% v = pv(1) + pv(2) * x + pv(3) * y + pv(4) * y^2 + pv(5) * xy

if(~CRISP)
    Zxq = zeros(3, 2);
    % first estimate is pretty straightforward:
    Zxq(1,1) = (pvq(2) + (HEIGHT/2) * pvq(5)) / Vyq; 
    Zxq(1,2) = f_var(Vyq);
    
    % then estimate w0:
    b = (puq(2) + 2 * puq(4) * (WIDTH/2) + puq(5) * (HEIGHT/2));
    rel_vert(1,1) = -b;
    rel_vert(1,2) = g_var(Vxq); % the closer to 0 the smaller the uncertainty
    rel_vert(2,1) = pvq(3) + 2 * pvq(4) * (HEIGHT/2) + pvq(5) * (WIDTH/2);
    rel_vert(2,2) = g_var(Vyq);
    w0 = ((1 /  rel_vert(1,2)) * rel_vert(1,1) + (1 /  rel_vert(2,2)) * rel_vert(2,1)) / ((1 /  rel_vert(1,2)) + (1 /  rel_vert(2,2)));
    var_w0 = 1 / (1 /  rel_vert(1,2) + 1 /  rel_vert(2,2));
    TTCq = 1/w0; 
    
    % second, third and fourth estimate directly depend on w0:
    Zxq(2,1) = (w0 + b) / Vxq;
    Zxq(2,2) = f_var(Vxq);
    Zxq(3,1) = -puq(4)/w0;
    Zxq(3,2) = var_w0;
    Zxq(4,1) = -pvq(4)/w0;
    Zxq(4,2) = var_w0;
    
    % combine all estimates of Zx:
    Zxq = combineEstimates(Zxq);
    
    % Now the same for zy:
    Zyq(1,1) = (puq(3) + (WIDTH/2) * puq(5)) / Vxq; 
    Zyq(1,2) = f_var(Vxq);
    
    % dv / dy = h + 2 * i * y + j * x
    h = pvq(3) + 2 * (HEIGHT/2) * pvq(4) + pvq(5) * (WIDTH / 2);
    Zyq(2,1) = (w0 - h) / Vyq;
    Zyq(2,2) = f_var(Vyq);
    Zyq(3,1) = puq(5) / w0; % is this correct?
    Zyq(3,2) = var_w0;
    Zyq(4,1) = -pvq(5) / w0;
    Zyq(4,2) = var_w0;
    
    % combine all estimates:
    Zyq = combineEstimates(Zyq);
    
   
else
    % best way is perhaps to start with the translational flow
    % depending on their magnitude, the relative vertical velocity can be
    % determined with a system of linear equations
    % if both the ventral flows are small, there are two equations
    % if only vx is small, vy can be used to estimate Zxq, which leads to
    % 1 + 2 = three equations for the relative vertical velocity.
    
    
    rel_vert_set = false;
    invalid_Zxq = true;
    if(abs(Vyq) > threshold_flow_translational)
        % estimate with the help of translational motion:
        % zx = vx / v0
        Zxq(1) = (pvq(2) + (HEIGHT/2) * pvq(5)) / Vyq;
        Zxq(2) = f_var(Vyq);
        invalid_Zxq = false;
    else
        % Vy close to zero, so this equation:
        % rel_vert = dv/dy - Vy * Zyq
        % now becomes independent of Zyq.
        rel_vert = pvq(3) + 2 * pvq(4) * (HEIGHT/2) + pvq(5) * (WIDTH/2);
        rel_vert_set = true;
        if(abs(rel_vert) > threshold_rel_vert)
            % there is considerable forward motion:
            Zxq1 = -pvq(5) / rel_vert;
            Zxq2 = -puq(4) / rel_vert;
            Zxq(1) = (Zxq1 + Zxq2) / 2;
            Zxq(2) = rel_vert;
            invalid_Zxq = false;
        else
            % there is little forward motion, but perhaps there is
            % horizontal motion:
            if(abs(Vxq) > threshold_flow_translational)
                Zxq(1) = (puq(2) + 2 * puq(4) * (WIDTH/2) + puq(5) * (HEIGHT/2)) / Vxq;
                Zxq(2) = f_var(Vxq);
                invalid_Zxq = false;
            else
                Zxq(1) = 0;
                Zxq(2) = 1E5;
            end
        end
    end
    
    
    invalid_Zyq = true;
    % estimate with the help of translational motion:
    if(abs(Vxq) > threshold_flow_translational)
        Zyq(1) = (puq(3) + puq(5) * (WIDTH/2)) / Vxq;
        Zyq(2) = f_var(Vxq);
        invalid_Zyq = false;
    else
        % Vxq close to zero, so this equation:
        % rel_vert = du/dx + Vxq * Zxq
        % now becomes independent of Zxq.
        rel_vert = puq(2) + 2 * puq(4) * (WIDTH/2) + puq(5) * (HEIGHT/2);
        rel_vert_set = true;
        if(abs(rel_vert) > threshold_rel_vert)
            % there is considerable forward motion:
            Zyq1 = -puq(5) / rel_vert;
            Zyq2 = -pvq(4) / rel_vert;
            Zyq(1) = (Zyq1 + Zyq2) / 2;
            Zyq(2) = rel_vert;
            invalid_Zyq = false;
        else
            % there is little forward motion, but perhaps there is
            % vertical motion:
            if(abs(Vyq) > threshold_flow_translational)
                Zyq(1) = (pvq(3) + 2 * (HEIGHT/2) * pvq(4) + pvq(5) * (WIDTH / 2)) / Vyq;
                Zyq(2) = f_var(Vyq);
                invalid_Zyq = false;
            else
                Zyq(1) = 0;
                Zyq(2) = 1E5;
            end
        end
    end
    
    if(~rel_vert_set)
        % we now have no less than 6 equations to solve for the relative
        % vertical velocity:
        rel_vert = zeros(6,1);
        rel_vert(1) = pvq(3) - Vyq * Zyq(1);
        rel_vert(2) = -puq(2) + Vxq * Zxq(1);
        rel_vert(3) = puq(5) / Zyq(1);
        rel_vert(4) = -pvq(4) / Zyq(1);
        rel_vert(5) = -pvq(5) / Zxq(1);
        rel_vert(6) = -puq(4) / Zxq(1);
        rel_vert = mean(rel_vert);
        rel_vert_set = true;
    end
    
    ux = puq(2) + 2 * puq(4) * (WIDTH/2) + puq(5) * (HEIGHT/2);
    vy = pvq(3) + 2 * pvq(4) * (HEIGHT/2) + pvq(5) * (WIDTH/2);
    Div = (ux+vy) / 2;
    TTCq = 1/Div;
    % TTCq = 1/rel_vert;
    
    if(invalid_Zxq)
        Zxq = [0, 0];
    end
    if(invalid_Zyq)
        Zyq = [0, 0];
    end
end


function Zx = combineEstimates(Zx)
% function Zx = combineEstimates(Zx)
% combine estimates as if they come from normal distributions:
% http://en.wikipedia.org/wiki/Sensor_fusion
n_estimates = size(Zx, 1);
nom = 0; den = 0;
for i = 1:n_estimates
    nom = nom + (1/Zx(i,2)) * Zx(i,1);
    den = den + 1/Zx(i,2);
end
Zx = [nom/ den, 1/den];

function y = f_var(x)
% function y = f_var(x)
if(abs(x) < 1)
    y = 1/x^2;
else
    y = 1 / abs(x);
end


function y = g_var(x, z)
% function y = g_var(x, z)
y = (20 * x) ^2;