function [X, Q] = findGoodFeatures(Im, window_size, ROI)
% function [X, Q] = findGoodFeatures(Im, window_size, ROI)
% Algorithm from: "Good features to track", by Shi and Tomasi
%
% - Im: Image in which to find features for optic flow tracking
% - window_size: window size that is the basis for the eigen values
% - ROI: region of interest in the image [min x, min y, max x, max y]
%
% (c) Guido de Croon, 2013

% only for debugging purposes:
verbose = false;
graphics = false;

% parameter processing:
if(~exist('window_size', 'var') || isempty(window_size))
    window_size = 5;
end
if(~exist('ROI', 'var') || isempty(ROI))
    ROI = [1 1 size(Im,2) size(Im,1)];
end


% get the size of the image
[H, W, C] = size(Im);

% Region Of Interest
min_x = max(window_size+1, ROI(1));
min_y = max(window_size+1, ROI(2));
max_x = min(W - window_size, ROI(3));
max_y = min(H - window_size, ROI(4));

% get the image gradient only of the ROI:
IX = zeros(H,W);
IY = zeros(H,W);
[Ix, Iy] = getGradient(Im(min_y:max_y, min_x:max_x), 'same');
IX(min_y:max_y, min_x:max_x) = Ix;
IY(min_y:max_y, min_x:max_x) = Iy;
Ix = IX;
Iy = IY;

% determine all the minimal eigenvalues of the "G"-matrices
MinEigVals = zeros(H - min_y - (H-max_y) + 1, W - min_x - (W-max_x) + 1);

for x = min_x : max_x
    for y = min_y : max_y
        G = getGPixel(Ix, Iy, x, y, window_size);
        eigs = sort(eig(G));
        MinEigVals(y-min_y+1,x-min_x+1) = eigs(1);
    end
end

if(graphics)
    h = figure();
    subplot(1,2,1); imshow(Im); title('Image');
    subplot(1,2,2); imagesc(MinEigVals); title('Minimum Eigen Values');
end

% Determine the maximum minimal eigenvalue:
max_min_eig = max(MinEigVals(:));
% threshold the points on the basis of this maximum:
threshold = 0.10;
RemainingPoints = zeros(H,W);
try
    RemainingPoints(min_y:max_y, min_x:max_x) = MinEigVals .* double(MinEigVals > threshold * max_min_eig);
catch err
    fprintf('Error\n');
end
if(graphics)
    hh = figure();
    subplot(1,2,1); imshow(Im); title('Image');
    subplot(1,2,2); imagesc(RemainingPoints); title('Thresholded Eigen Values');
end
% clean up remaining points, removing those that are close:
n_h_size = 5;
N_HOOD = zeros(2*n_h_size+1);%[0 0 0; 0 1 0; 0 0 0];
N_HOOD(n_h_size+1, n_h_size+1) = 1;

for x = min_x : max_x
    for y = min_y : max_y
        if(RemainingPoints(y,x) > 0)
            % clean up the neighbors:
            RemainingPoints(y-n_h_size:y+n_h_size,x-n_h_size:x+n_h_size) = RemainingPoints(y-n_h_size:y+n_h_size,x-n_h_size:x+n_h_size) .* N_HOOD;
        end
    end
end

% the remaining points are good for tracking:
[y,x] = find(RemainingPoints > 0);
RP = RemainingPoints(:);
Q = RP((x-1)*size(RemainingPoints,1)+y);
X = [x, y];

if(graphics)
    hhh = figure();
    subplot(1,2,1); imshow(Im); title('Image');
    subplot(1,2,2); imagesc(RemainingPoints); title('Thresholded Eigen Values');
    pause;
    close(h); close(hh); close(hhh);
end
