function [FoE, TTC, Vx, Vy, Zx, Zy, pu, pv, points, erroneous, inliers, errors] = get_TTI_from_flow_Image_Kalman(Im1, Im2, graphics, pu, pv, FoE, points)
% function [FoE, TTC, Vx, Vy, Zx, Zy, pu, pv, points, erroneous, inliers, errors] = get_TTI_from_flow_Image_Kalman(Im1, Im2, graphics, pu, pv, FoE, points)
%
% For two given images determines the focus of expansion and
% time-to-contact. This Kalman version tracks points over multiple images,
% estimating their image location and velocity over time.
%
% parameters:
% - Im1: previous image
% - Im2: current image
% - graphics: whether to show all intrinsics of the TTC determination.
% - pu, pv: parameters of the planes representing (du/dx, du/dy), (dv/dx,
%           dv/dy) or the quadratic parametrization
% - FoE: focus of expansion
% - points: tracked points 
%
% outputs:
% - FoE: focus of expansion
% - TTC: time-to-contact estimate
% - Vx, Vy: translational flow
% - Zx, Zy: slope of the surface + certainty
% - pu, pv: see above
% - points: updated tracked points 
% - erroneous: whether the estimate is erroneous
% - inliers: ratio of inliers for the flow plane fit
%
% The code has not been optimized for speed.
% (c) Guido de Croon, 10-12-2011

% initialize outputs:
if(~exist('pu', 'var'))
    FoE = []; pu = []; pv = []; points = [];
end
TTC = []; Vx = []; Vy = []; Zx = [0,0]; Zy = [0,0];
erroneous = 0; inliers = 1; errors = 0;

verbose = true;

if(~exist('graphics', 'var') || isempty(graphics))
    graphics = false;
end

% whether to use the divergence of the flow for estimating TTC:
DIVERGENCE_PLANE = 1;
DIVERGENCE_QUADRATIC = 2;
ESTIMATE_FOE = 3;
TTC_METHOD = DIVERGENCE_QUADRATIC;

% Kalman parameters:
% the state is: [x,y,u,v] (last two represent a flow vector)
V_init = 0;
HIGH_MEASUREMENT_NOISE = false; % was false for optimal landing study
if(exist('pu', 'var') && ~isempty(pu))
    if(TTC_METHOD == DIVERGENCE_QUADRATIC)
        filteredDivergence = pu(2) + pv(3);
    else
        filteredDivergence = pu(1) + pv(2);
    end
    filteredTTC = 2 / filteredDivergence;
    THRESHOLD_FRAMES = 100;
    if(filteredTTC < THRESHOLD_FRAMES)
        HIGH_MEASUREMENT_NOISE = false;
    end
end
if(HIGH_MEASUREMENT_NOISE)
    % trust motion model a lot - reducing noise
    pixel_var = 1;
    flow_var = 5;
    movement_var = 0.1;
else
    % trust measurements a lot - reducing delay
    pixel_var = 0.1;
    flow_var = 0.5;
    movement_var = 5;
end
P_init = diag([pixel_var, pixel_var, flow_var, flow_var]);
Q_init = diag([movement_var, movement_var, movement_var, movement_var]);
R_init = diag([pixel_var, pixel_var, flow_var, flow_var]);
% models the update equations for the prediction step:
F = [1 0 1 0; 0 1 0 1; 0 0 1 0; 0 0 0 1];

% whether to use the previous estimate of the flow field to predict the
% feature points in the next image:
GUESSING_FLOW = true; %true for the optimal landing 
if(GUESSING_FLOW)
    SEA_Colors = zeros(3);
    SEA_Colors(1,:) = [1 1 1];
    SEA_Colors(2,:) = [0 1 0];
    SEA_Colors(3,:) = [1 1 1];
end
% this guessing of flow is only possible in case the parameters pu, pv are present:
% if(~DIVERGENCE || ~exist('pu', 'var') || ~exist('pv', 'var'))
if(~exist('pu', 'var') || ~exist('pv', 'var'))
    GUESSING_FLOW = false;
end

% whether to find features in the entire image every image, or use
% information from previous tranlsation / FoE to only search in areas
% without points:
EFFICIENT_FEATURE_FINDING = false; % there will be errors for quadratic flow estimation here!
% if not using efficient feature finding, search for new features if there
% are fewer than MIN_FEATURES
MIN_FEATURES = 20; %12;
% cap off the number of features at MAX_FEATURES for computational reasons:
MAX_FEATURES = 50;

% directory containing findF, etc.:
% addpath('C:/Research/3D/MVG/');
if(~exist('FoE', 'var'))
    FoE = [];
end
TTC = []; Vx = []; Vy = []; Y =[];

% whether to only accept flow vectors that arrive at the same point when
% traced back to the old image:
FLOW_2_WAY = true;

% whether to use the graphics of the LKT script:
LKT_graphics = false;

% image resize for finding the right features
IM_RESIZE = 1;%0.5; 

% CELLS is the number of separate image regions in which to find
% points, so that they are distributed spatially. Should always be
% a squared number (1, 4, 9, etc.)

% !!! Does this work well together with efficient feature finding? !!!

CELLS = 1; % 4 for optimal landing

% get height and width:
HEIGHT = size(Im1, 1);
WIDTH = size(Im1, 2);

if(EFFICIENT_FEATURE_FINDING && exist('pu', 'var') && exist('pv', 'var') && exist('FoE', 'var') && ~isempty(FoE))
    % serious errors in this part, main choice to make:
    % does getFeatures add points to X, or return only new points?
    
    % take into account translation and FoE in searching for new features:
    X = points.X;
    X = getFeaturesToTrackEfficient(X, transformImage(Im1), pu, pv, FoE, TTC_METHOD);
    % add points:
    n_old_points = size(points.X, 1);
    n_added = size(X,1) - n_old_points;
    if(n_added > 0)
        points.X = [points.X; X(n_old_points+1:end, :)];
        points = initializeAddedPoints(points, V_init, P_init, Q_init, R_init);
    end
end

erroneous = false;

if(exist('points', 'var') && ~isempty(points))
    X = points.X;
end

if(~GUESSING_FLOW || ~exist('pu', 'var') || isempty(pu)) % this pu-stuff needs to be replaced if we start also using FoE / Delaunay
    % detect the flow:
    % if(~exist('X', 'var') || isempty(X) || (~EFFICIENT_FEATURE_FINDING && size(X,1) < MIN_FEATURES))
    if(~exist('X', 'var') || isempty(X))
        % no points yet, get some with flow:
        %         window_size = 5;
        %         ROI = [1 1 WIDTH HEIGHT];
        %         [X, Q] = findGoodFeatures(imresize(transformImage(Im1), IM_RESIZE), window_size, ROI, CORNER_METHOD);
        %         X = X ./ IM_RESIZE;
        
        ImPoints = imresize(transformImage(Im1), IM_RESIZE);
        window_size = 5;
       
        if(CELLS == 1)
            ROI = [1 1 WIDTH HEIGHT];
            [X, Q] = findGoodFeatures(ImPoints, window_size, ROI);
        else
            % we need to still test this for FAST9 and ACTIVE_CORNERS:
            WRES = size(ImPoints,2);
            HRES = size(ImPoints,1);
            XX = cell(CELLS,1);
            n_hor = sqrt(CELLS);
            n_pix_h = floor(WRES / n_hor);
            n_pix_v = floor(HRES / n_hor);
            it = 1;
            for h = 1:n_hor
                for v = 1:n_hor
                    ROI = [1+n_pix_h*(h-1), 1+n_pix_v*(v-1), h*n_pix_h, v*n_pix_v];
                    [XX{it}, Q] = findGoodFeatures(ImPoints, window_size, ROI);
                    it = it + 1;
                end
            end
            X = [];
            for c = 1:CELLS
                X = [X; XX{c}];
            end
        end
        X = X * (1 / IM_RESIZE);
        X = limitNumberOfFeatures(X, MAX_FEATURES);
        [X,Y,V,final_errors] = pyramidLKT(Im1, Im2, [], [], 2, LKT_graphics, X, []); % window_size
                
        % store the information in points, then initialize:
        points.X = Y;
        points.V = V;
        points.P = [];
        points.Q = [];
        points.R = [];
        points = initializeAddedPoints(points, V_init, P_init, Q_init, R_init);
    else
        % first run the Kalman prediction step:
        n_points = size(X,1);
        pre_state = cell(n_points,1);
        preP = cell(n_points,1);
        S = cell(n_points,1);
        K = cell(n_points,1);
        for p = 1:n_points
            % determine the pre-values:
            pre_state{p} = [points.X(p,:) + points.V(p,:), points.V(p, :)]';
            preP{p} = F * points.P{p}' * F' + points.Q{p};
            S{p} = preP{p} + points.R{p};
            K{p} = preP{p} * pinv(S{p});
        end
        
        % determine flow:
        [X,Y,V,final_errors] = pyramidLKT(Im1, Im2, [], [], 2, LKT_graphics, X, []);
        
        % update the variables on the basis of the observations:
        % Kalman correction step:
        for p = 1:n_points
            % correct on the basis of the observation:
            observation = [Y(p,:), V(p,:)]';
            innovation = observation - pre_state{p};
            state = pre_state{p} + K{p} * innovation;
            points.P{p} = (eye(4) - K{p}) * preP{p};
            points.X(p,:) = state(1:2)';
            points.V(p,:) = state(3:4)';
        end
        
        if(~EFFICIENT_FEATURE_FINDING && size(X,1) < MIN_FEATURES)
            % newX could have the same points as X:
            ROI = [1 1 WIDTH HEIGHT]; window_size = 5; ImPoints = imresize(transformImage(Im1), IM_RESIZE);
            [newX, Q] = findGoodFeatures(ImPoints, window_size, ROI);
            newX = newX * (1 / IM_RESIZE);
            newX = limitNumberOfFeatures(newX, MAX_FEATURES - size(X,1));
            
            [newX,newY,newV,newfinal_errors] = pyramidLKT(Im1, Im2, [], [], 2, LKT_graphics, newX, []);
            inds = find(~isnan(newY(:,1)));
            newY = newY(inds,:); newV = newV(inds,:);
            
            % store the information in points, then initialize:
            newpoints.X = newY;
            newpoints.V = newV;
            newpoints.P = [];
            newpoints.Q = [];
            newpoints.R = [];
            newpoints = initializeAddedPoints(newpoints, V_init, P_init, Q_init, R_init);
            points = joinPoints(points, newpoints);
            X = points.X;
            V = points.V;
            Y = X + V;
        end
        
    end
else
    % 1. detect new points in Im1
    % 2. insert their coordinates in the previously estimated "plane",
    %    representing the flow at that point in the last image
    % 3. send these flow vectors as V_guess to pyramid_LKT
    
    new_points = false;
    
    % 1:
    if(~exist('X', 'var') || isempty(X))% || (~EFFICIENT_FEATURE_FINDING && size(X,1) < MIN_FEATURES))
        new_points = true;
        ImPoints = imresize(transformImage(Im1), IM_RESIZE);
        window_size = 5;
        if(CELLS == 1)
            ROI = [1 1 WIDTH HEIGHT];
            [X, Q] = findGoodFeatures(ImPoints, window_size, ROI);
        else
            % we need to still test this for FAST9 and ACTIVE_CORNERS:
            WRES = size(ImPoints,2);
            HRES = size(ImPoints,1);
            XX = cell(CELLS,1);
            n_hor = sqrt(CELLS);
            n_pix_h = floor(WRES / n_hor);
            n_pix_v = floor(HRES / n_hor);
            it = 1;
            for h = 1:n_hor
                for v = 1:n_hor
                    ROI = [1+n_pix_h*(h-1), 1+n_pix_v*(v-1), h*n_pix_h, v*n_pix_v];
                    [XX{it}, Q] = findGoodFeatures(ImPoints, window_size, ROI);
                    it = it + 1;
                end
            end
            X = [];
            for c = 1:CELLS
                X = [X; XX{c}];
            end
        end
        X = X * (1 / IM_RESIZE);
    elseif(~EFFICIENT_FEATURE_FINDING && size(X,1) < MIN_FEATURES)
         % newX could have the same points as X:
         ROI = [1 1 WIDTH HEIGHT]; window_size = 5; ImPoints = imresize(transformImage(Im1), IM_RESIZE);
         [newX, Q] = findGoodFeatures(ImPoints, window_size, ROI);
         newX = newX * (1 / IM_RESIZE);
         newX = limitNumberOfFeatures(newX, MAX_FEATURES - size(X,1));
         % store the information in points, then initialize:
         newpoints.X = newX;
         newpoints.V = zeros(length(newX),2);
         newpoints.P = [];
         newpoints.Q = [];
         newpoints.R = [];
         newpoints = initializeAddedPoints(newpoints, V_init, P_init, Q_init, R_init);
         points = joinPoints(points, newpoints);
         X = points.X;
    end
    
    if(isempty(X))
        TTC = 0;
        Vx = 0;
        Vy = 0;
        erroneous = true;
        return;
    end
    
    % 2:
    if(TTC_METHOD == DIVERGENCE_PLANE)
        V_X = pu(1) .* X(:,1) + pu(2) .* X(:,2) + pu(3);
        V_Y = pv(1) .* X(:,1) + pv(2) .* X(:,2) + pv(3);
    else
        V_X = pu(1) + pu(2) .* X(:,1) + pu(3) .* X(:,2) + pu(4) .* X(:,1) .* X(:,1) + pu(5) .* X(:,1) .* X(:,2);
        V_Y = pv(1) + pv(2) .* X(:,1) + pv(3) .* X(:,2) + pv(4) .* X(:,2) .* X(:,2) + pv(5) .* X(:,1) .* X(:,2);
    end
    V_guess = [V_X, V_Y];
    
    % 3:
    [X,Y,V,final_errors] = pyramidLKT(Im1, Im2, [], [], 2, LKT_graphics, X, V_guess);
    normal_inds = find(~isnan(V(:,1)));
    X = X(normal_inds, :);
    Y = Y(normal_inds, :);
    V = V(normal_inds, :);
    V_guess = V_guess(normal_inds, :);
    final_errors = final_errors(normal_inds, 1);
    
    if(new_points)
        % store the information in points, then initialize:
        points.X = Y;
        points.V = V;
        points.P = [];
        points.Q = [];
        points.R = [];
        points = initializeAddedPoints(points, V_init, P_init, Q_init, R_init);
    else
        
        % only keep points for which a measurement was made (this can be
        % changed to give a point a chance to be "refound"
        points.X = points.X(normal_inds, :);
        points.V = points.V(normal_inds, :);
        points.P = points.P(normal_inds);
        points.Q = points.Q(normal_inds);
        points.R = points.R(normal_inds);        
        
        % first run the Kalman prediction step:
        n_points = size(X,1);
        pre_state = cell(n_points,1);
        preP = cell(n_points,1);
        S = cell(n_points,1);
        K = cell(n_points,1);
        for p = 1:n_points
            % determine the pre-values:
            pre_state{p} = [points.X(p,:) + points.V(p,:), points.V(p, :)]';
            preP{p} = F * points.P{p}' * F' + points.Q{p};
            S{p} = preP{p} + points.R{p};
            K{p} = preP{p} * pinv(S{p});
        end
        
        % update the variables on the basis of the observations:
        % Kalman correction step:
        for p = 1:n_points
            % correct on the basis of the observation:
            observation = [Y(p,:), V(p,:)]';
            innovation = observation - pre_state{p};
            state = pre_state{p} + K{p} * innovation;
            points.P{p} = (eye(4) - K{p}) * preP{p};
            points.X(p,:) = state(1:2)';
            points.V(p,:) = state(3:4)';
        end
    end
end

n_matches = size(X,1);

if(graphics)
    h = figure();
    if(FLOW_2_WAY)
        subplot(1,2,1);
    end
    showFlow(Im1, Im2, X, Y);
    if(GUESSING_FLOW && exist('pu', 'var') && exist('V_guess', 'var'))
        hold on;
        showFlow(Im1, Im2, X, X+V_guess, SEA_Colors);
    end
end

if(FLOW_2_WAY && ~isempty(Y)) 
    % how effective would error / quality filtering be?
    % send the points in the new image to the optic flow procedure with
    % the images reversed. YY should be equal to X
    if(~GUESSING_FLOW || ~exist('pu', 'var') || isempty(pu))
        [XX,YY,VV,final_errors] = pyramidLKT(Im2, Im1, [], [], 2, LKT_graphics, Y);
    else
        [XX,YY,VV,final_errors] = pyramidLKT(Im2, Im1, [], [], 2, LKT_graphics, Y, -V_guess);
    end
    D = sum(abs(YY - X), 2);
    % (pixel_threshold/2) deviation per flow component max:
    pixel_threshold = 2;
    keep_inds = D < pixel_threshold;
    % keep only the accurate ones:
    X = X(keep_inds, :);
    n_matches = size(X,1);
    Y = Y(keep_inds, :);
    V = V(keep_inds, :);
    
    % update the points, only retaining successful matches:
    points = selectPoints(points, keep_inds);
    
    if(graphics)
        figure(h);
        subplot(1,2,2);
        showFlow(Im1, Im2, X, Y);
        if(GUESSING_FLOW && exist('pu', 'var') && exist('V_guess', 'var'))
            V_guess = V_guess(keep_inds,:);
            hold on;
            showFlow(Im1, Im2, X, X+V_guess, SEA_Colors);
        end
    end
end


if(TTC_METHOD == DIVERGENCE_QUADRATIC)
    % We want to fit a plane through the flow vectors
    
    % first the slope for u (flow_x):
    % in principle, we could restrict the slope in the y-direction to
    % be zero:
    n_matches = size(points.X,1);
    
    % quadratic fit:
    [puq, pvq, inliersv, inliersu, error_u, error_v, rank_deficient] = estimateFlowQuadratic(X, V, n_matches, WIDTH, HEIGHT);
    
    if(~exist('puq', 'var') || isempty(puq) || rank_deficient)
        % set TTC to 0, so that one can know that the estimate is
        % erroneous:
        TTC = 0;
        FoE = [-128,-128];
        Vx = 0;
        Vy = 0;
        Zx = [0, 0];
        Zy = [0, 0];
        pu = [];
        pv = [];
        inliers = 0;
        errors = -1;
        erroneous = true;
        return;
    else
        erroneous = false;
    end
    
    if(verbose)
        fprintf('u = %f + %f x + %f y + %f x^2 + %f xy\n', puq(1), puq(2), puq(3), puq(4), puq(5));
        fprintf('v = %f + %f x + %f y + %f y^2 + %f xy\n', pvq(1), pvq(2), pvq(3), pvq(4), pvq(5));
    end
    
    % extract ventral flow, time-to-contact, slopes, etc:
    [Vxq, Vyq, TTCq, Zxq, Zyq] = extractFlowFromQuadraticEquation(puq, pvq, WIDTH, HEIGHT);
    
    % transform to output form:
    pu = puq; pv = pvq; Vx = Vxq; Vy = Vyq; Zx = Zxq; Zy = Zyq; TTC = TTCq;
    inliers = (inliersv + inliersu) / 2;
    errors = (error_u + error_v) / 2;
    if(verbose)
        fprintf('Quadratic fit: TTC %f, Vx = %f, Vy = %f\n', TTCq, Vxq, Vyq);
        %         if(~isempty(Zyq))
        %             fprintf('Zy = %f\n', Zyq);
        %         end
        %         if(~isempty(Zxq))
        %             fprintf('Zx = %f\n', Zxq);
        %         end
    end
    
    if(graphics)
        grid_step = 20;
        grid_x = round(grid_step/2)-0.5:grid_step:WIDTH;
        n_x = length(grid_x);
        grid_y = round(grid_step/2)-0.5:grid_step:HEIGHT;
        n_y = length(grid_y);
        GRID_X = repmat(grid_x', [n_y,1]);
        GRID_X = GRID_X(:);
        GRID_Y = repmat(grid_y, [n_x, 1]);
        GRID_Y = GRID_Y(:);
        zu = puq(1) + puq(2) .* GRID_X + puq(3) .* GRID_Y + puq(4) .* GRID_X .* GRID_X + puq(5) .* GRID_X .* GRID_Y;
        zv = pvq(1) + pvq(2) .* GRID_X + pvq(3) .* GRID_Y + pvq(4) .* GRID_Y .* GRID_Y + pvq(5) .* GRID_X .* GRID_Y;
        ZU = reshape(zu, [n_x, n_y])';
        ZV = reshape(zv, [n_x, n_y])';
        % show the flow in 3D:
        figure();
        plot3(X(:,1), X(:,2), V(:,1), 'x', 'Color', 'blue');
        hold on;
        plot3(X(:,1), X(:,2), V(:,2), 'x', 'Color', 'red');
        surface(grid_x, grid_y, ZU);
        surface(grid_x, grid_y, ZV);
        legend('flow x', 'flow y');
        ylabel('Y'); xlabel('X'); zlabel('Flow');
        
        % show the plane fit:
        figure();
        subplot(1,2,1);
        imagesc(ZU);
        title('du(x,y) / dx');
        subplot(1,2,2);
        imagesc(ZV);
        title('dv(x,y) / dx');
        
        % graphics for in article:
        % 1) flow:
        h_flow = figure();
        showFlow(Im1, Im1, X, Y); % twice Im1
        % 2) use parameters puq to create flow field:
        figure();
        showFlow(Im1, Im2, [GRID_X, GRID_Y], [GRID_X+zu, GRID_Y]);
        xc = (WIDTH/2); yc = (HEIGHT/2);
        ventral_x = puq(1) + puq(2) .* xc + puq(3) .* yc + puq(4) .* xc^2 + puq(5) .* yc .* xc;
        %arrow([xc, yc], [xc + ventral_x, yc], 'Length', 3, 'EdgeColor', [1 1 1], 'FaceColor', [1 1 1]);
        plot(xc+1, yc+1, 'o', 'Color', [1 1 1], 'MarkerSize', 8);
        % 3) use parameters pvq to create flow field:
        figure();
        showFlow(Im1, Im2, [GRID_X, GRID_Y], [GRID_X, GRID_Y+zv]);
        ventral_y = pvq(1) + pvq(2) .* xc + pvq(3) .* yc + pvq(4) .* yc^2 + pvq(5) .* yc .* xc;
        %arrow([xc, yc], [xc, yc + ventral_y], 'Length', 3, 'EdgeColor', [1 1 1], 'FaceColor', [1 1 1]);
        plot(xc+1, yc+1, 'o', 'Color', [1 1 1], 'MarkerSize', 8);
        
        figure(h_flow);
        
        % replace this with some valid formulas:
        FoE = [WIDTH/2, HEIGHT/2];
        
    end
elseif(TTC_METHOD == DIVERGENCE_PLANE)
    % ***********
    % Linear fit:
    % ***********
    
    [pu, pv, inliersv, inliersu, error_u, error_v, rank_deficient] = estimateFlowPlanes(points.X, points.V, n_matches, WIDTH, HEIGHT);
    
    if(isempty(pu) || isempty(pv) || rank_deficient)
        % not enough points, set them as if TTC = 1:
        % pu = [1 0 0];
        % pv = [0 1 0];
        % but set TTC to 0, so that one can know that the estimate is
        % erroneous:
        TTC = 0;
        FoE = [-128,-128];
        Vx = 0;
        Vy = 0;
        Zx = [0, 0];
        Zy = [0, 0];
        pu = [];
        pv = [];
        inliers = 0;
        errors = -1;
        erroneous = true;
        return;
    else
        if(verbose)
            fprintf('u = %f x + %f y + %f\n', pu(1), pu(2), pu(3));
            fprintf('v = %f x + %f y + %f\n', pv(1), pv(2), pv(3));
        end
        
        % Divergence = du(x,y) / dx + dv(x,y) / dy
        Divergence = pu(1) + pv(2);
        TTC = 2 / Divergence;
        inliers = (inliersu + inliersv) / 2;
        errors = (error_u + error_v) / 2;
    end
        
    % Ventral flow is the value of the planes in the center of the image:
    Vx = -(pu(3) + (WIDTH/2) * pu(1) + (HEIGHT/2) * pu(2));
    Vy = -(pv(3) + (WIDTH/2) * pv(1) + (HEIGHT/2) * pv(2));
    
    % if Vx, Vy large enough, estimate Zx, Zy:
    threshold_slope = 1.0;
    eta = 0.002;
    if(abs(pv(2)) < eta && abs(Vy) < threshold_slope && abs(Vx) >= 2*threshold_slope)
        % in this case, Zx = pu(1) / Vx
        Zx(1) = pu(1) / Vx;
        Zx(2) = Vx;
        fprintf('No forward / vertical movement: Zx = %f\n', Zx(1));
    elseif(abs(Vy) >= 2*threshold_slope)
        Zx(1) = pv(1) / Vy;
        Zx(2) = Vy;
        if(verbose)
            fprintf('Zx = %f\n', Zx);
        end
    else
        Zx = [0, 0];
    end
    if(abs(pu(1)) < eta && abs(Vx) < threshold_slope && abs(Vy) >= 2*threshold_slope)
        % in this case, Zy = pv(3) / Vy
        Zy(1) = pv(2) / Vy;
        Zy(2) = Vy;
        fprintf('No forward / horizontal movement: Zy = %f\n', Zy(1));
    elseif(abs(Vx) >= 2*threshold_slope)
        Zy(1) = pu(2) / Vx;
        Zy(2) = Vx;
        if(verbose)
            fprintf('Zy = %f\n', Zy);
        end
    else
        Zy = [0, 0];
    end
    
    % FoE (1)
    % on the basis of the vectors:
    % [fx, fy] = findFoE(X,Y,V,HEIGHT,WIDTH);
    % FoE(j,:) = [-fx,-fy];
    % FoEIm = [WIDTH / 2 - fx, HEIGHT/ 2 - fy];
    
    % FoE (2)
    % on the basis of the planes:
    denominator = (pv(1) * pu(2) - pu(1) * pv(2));
    if(abs(denominator) >  eps)
        FoEIm(1,1) = (pu(3) * pv(2) - pv(3) * pu(2)) / denominator;
    else
        FoEIm(1,1) = 0;
    end
    denominator = pu(2);
    if(abs(denominator) >  eps)
        FoEIm(1,2) = -(pu(1) * FoEIm(1,1) + pu(3)) / denominator;
    else
        FoEIm(1,2) = 0;
    end
    FoE = FoEIm - [WIDTH/2, HEIGHT/2];
    
    if(verbose) 
        fprintf('TTC = %f\n', TTC);
        fprintf('Vx = %f, Vy = %f\n', Vx, Vy);
        fprintf('FoE = (%f,%f)\n', FoEIm(1,1), FoEIm(1,2));
    end
    if(graphics)
        grid_step = 20;
        grid_x = round(grid_step/2)-0.5:grid_step:WIDTH;
        n_x = length(grid_x);
        grid_y = round(grid_step/2)-0.5:grid_step:HEIGHT;
        n_y = length(grid_y);
        GRID_X = repmat(grid_x', [n_y,1]);
        GRID_X = GRID_X(:);
        GRID_Y = repmat(grid_y, [n_x, 1]);
        GRID_Y = GRID_Y(:);
        zu = pu(1) .* GRID_X + pu(2) .* GRID_Y + pu(3);
        zv = pv(1) .* GRID_X + pv(2) .* GRID_Y + pv(3);
        ZU = reshape(zu, [n_x, n_y])';
        ZV = reshape(zv, [n_x, n_y])';
        % show the flow in 3D:
        figure();
        plot3(X(:,1), X(:,2), V(:,1), 'x', 'Color', 'blue');
        hold on;
        plot3(X(:,1), X(:,2), V(:,2), 'x', 'Color', 'red');
        surface(grid_x, grid_y, ZU);
        surface(grid_x, grid_y, ZV);
        legend('flow x', 'flow y');
        ylabel('Y'); xlabel('X'); zlabel('Flow');
        
        % show the plane fit:
        figure();
        subplot(1,2,1);
        imagesc(ZU);
        title('du(x,y) / dx');
        subplot(1,2,2);
        imagesc(ZV);
        title('dv(x,y) / dx');
        
        % graphics for in article:
        % 1) flow:
        h_flow = figure();
        showFlow(Im1, Im1, X, Y); % twice Im1
        % 2) use parameters pu to create flow field:
        figure();
        showFlow(Im1, Im2, [GRID_X, GRID_Y], [GRID_X+zu, GRID_Y]);
        xc = (WIDTH/2); yc = (HEIGHT/2);
        ventral_x = pu(1) .* xc + pu(2) .* yc + pu(3);
        %arrow([xc, yc], [xc + ventral_x, yc], 'Length', 3, 'EdgeColor', [1 1 1], 'FaceColor', [1 1 1]);
        plot(xc+1, yc+1, 'o', 'Color', [1 1 1], 'MarkerSize', 8);
        % 3) use parameters pv to create flow field:
        figure();
        showFlow(Im1, Im2, [GRID_X, GRID_Y], [GRID_X, GRID_Y+zv]);
        ventral_y = pv(1) .* xc + pv(2) .* yc + pv(3);;
        %arrow([xc, yc], [xc, yc + ventral_y], 'Length', 3, 'EdgeColor', [1 1 1], 'FaceColor', [1 1 1]);
        plot(xc+1, yc+1, 'o', 'Color', [1 1 1], 'MarkerSize', 8);
        
        %pause;
        %close all;
        
        figure(h_flow);
        
    end
    
elseif(TTC_METHOD == ESTIMATE_FOE)
    
    if(size(X,1) < 5)
        % too few features:
        FoE = [0,0];
        FoEIm = [WIDTH / 2, HEIGHT/ 2];
        TTC = 0;
        Vx = 0; Vy = 0;
        inliers = 0;
        errors = -1;
        erroneous = true;
        return;
    end
    
    % find the focus of expansion by a minimum least squares solution:
    [fx, fy] = findFoE(X,Y,V,HEIGHT,WIDTH);
    FoE = [-fx,-fy];
    FoEIm = [WIDTH / 2 - fx, HEIGHT/ 2 - fy];
    
    % find the time to contact:
    [TTC, TTC_distr] = determineTTC(X,Y,V,HEIGHT,WIDTH,FoEIm);
    
    % fit a plane - abusing the existing function for flow estimation:
    [pars1, pars2, inliers1, inliers2, error_1, error_2, rank_deficient] = estimateFlowPlanes(X, [TTC_distr, TTC_distr], n_matches, WIDTH, HEIGHT);
    
    % necessary:
    if(isempty(inliers1))
        inliers = 0;
        errors = -1;
        erroneous = true;
    else
        inliers = inliers1;
        errors = error_1;
        erroneous = false;
        % wrong approximation:
        Vx = median(V(:,1));
        Vy = median(V(:,2));
        
        Zx(1) = (pars1(1) + pars2(1)) / 2;
        Zy(1) = (pars1(2) + pars2(2)) / 2;
    end
    pu = []; pv = [];
    
    if(verbose)
        fprintf('FoE = (%f, %f), TTC = %f\n', FoE(1), FoE(2), TTC);
    end
    
    if(graphics)
        plot(FoEIm(1), FoEIm(2), 'o', 'MarkerSize', 10, 'LineWidth', 3, 'Color', [1 0 1]);
        %         GTFoEIm = [WIDTH/2 + GTFoEX(1), HEIGHT/2 + GTFoEY(1)];
        %         plot(GTFoEIm(1), GTFoEIm(2), 'o', 'MarkerSize', 10, 'LineWidth', 3, 'Color', [0 1 0]);
        %         title(['FoE = (' num2str(FoE(j,1)) ', ' num2str(FoE(j,2)) '), GTFoE = (' num2str(GTFoEX(1)) ', ' num2str(GTFoEY(1)) ')']);
        %         pause();
        %         close all;
    end
end

function surfaces = getSurfaces(X, tri)
% function surfaces = getSurfaces(X, tri)
n_triangles = size(tri,1);
surfaces = zeros(n_triangles, 1);
for t = 1:n_triangles
    triangle = tri(t,:);
    x_1 = [X(triangle(1), :), 0];
    x_2 = [X(triangle(2), :), 0];
    x_3 = [X(triangle(3), :), 0];
    v_12 = x_2 - x_1;
    v_13 = x_3 - x_1;
    surfaces(t) = norm(cross(v_12, v_13));
end

function showFlow(Im1, Im2, X, Y, SEA_Colors)
% function showFlow(Im1, Im2, X, Y, SEA_Colors)

CROSSES = false;

if(~exist('SEA_Colors', 'var') || isempty(SEA_Colors))
    SEA_Colors = zeros(3);
    SEA_Colors(1,:) = [0 0 1];
    SEA_Colors(2,:) = [1 0 0];
    SEA_Colors(3,:) = [1 1 1];
end

n_points = size(X,1);
imshow(0.5 * Im1 + 0.5 * Im2);
hold on;
if(CROSSES)
    plot(X(:,1), X(:,2), 'x', 'Color', SEA_Colors(1,:));
    plot(Y(:,1), Y(:,2), 'x', 'Color', SEA_Colors(2,:));
end
if(exist('arrow.m', 'file'))
    % plot arrows for the flow:
    for p = 1:n_points
        if(~isnan(Y(p,1)))
            arrow(X(p,:), Y(p,:), 'Length', 3, 'EdgeColor', SEA_Colors(3,:), 'FaceColor', SEA_Colors(3,:));
        end
    end
else
    error('Please download arrow.m - see README.txt.');
end

function points = joinPoints(old_points, new_points)
% function points = joinPoints(old_points, new_points)

SUPPRESS = true;
if(SUPPRESS && size(new_points.X, 1) >= 1)
    suppression_range = 5;
    remove_vec = zeros(size(new_points.X,1),1);
    for o = 1:size(old_points.X, 1)
        for n = 1:size(new_points.X, 1)
            dis = pdist([old_points.X(o,:); new_points.X(n,:)]);
            if(dis < suppression_range)
                remove_vec(n) = 1;
            end
        end
    end
    keep_inds = find(remove_vec == 0);
    new_points.X = new_points.X(keep_inds, :);
    new_points.V = new_points.V(keep_inds, :);
    new_points.P = new_points.P(keep_inds, :);
    new_points.Q = new_points.Q(keep_inds, :);
    new_points.R = new_points.R(keep_inds, :);
end

points.X = [old_points.X; new_points.X];
points.V = [old_points.V; new_points.V];
points.P = [old_points.P; new_points.P];
points.Q = [old_points.Q; new_points.Q];
points.R = [old_points.R; new_points.R];


function points = initializeAddedPoints(points, V_init, P_init, Q_init, R_init)
% function points = initializeAddedPoints(points, V_init, P_init, Q_init, R_init)

% it can happen that the optic flow vectors (velocities) are already added:
n_points = size(points.X,1);
if(size(points.V,1) < size(points.X,1))
    n_added = size(points.X,1) - size(points.V,1);
    points.V = [points.V; V_init .* ones(n_added, 2)];
end

if(size(points.P,1) < size(points.X,1))
    n_added = size(points.X,1) - size(points.P,1);
    P_add = cell(n_added, 1);
    Q_add = cell(n_added, 1);
    R_add = cell(n_added, 1);
    
    for a = 1:n_added
        P_add{a} = P_init;
        Q_add{a} = Q_init;
        R_add{a} = R_init;
    end
    
    points.P = [points.P; P_add];
    points.Q = [points.Q; Q_add];
    points.R = [points.R; R_add];
end

function points = selectPoints(points, keep_inds)
% function points = selectPoints(points, keep_inds)
points.X = points.X(keep_inds, :);
points.V = points.V(keep_inds, :);
points.P = points.P(keep_inds);
points.Q = points.Q(keep_inds);
points.R = points.R(keep_inds);


function X = limitNumberOfFeatures(X, MAX_FEATURES)
% function X = limitNumberOfFeatures(X, MAX_FEATURES)
nf = size(X,1);
if(nf > MAX_FEATURES)
    inds = randperm(nf);
    X = X(inds(1:MAX_FEATURES), :);
end



