function [X,Y,V,final_errors,Q] = pyramidLKT(Im1, Im2, window_size, ROI, pyramid_levels, graphics, X, V_guess)
% function [X,Y,V,final_errors,Q] = pyramidLKT(Im1, Im2, window_size, ROI, pyramid_levels, graphics, X, V_guess)
% Implementation of the pyramidal Lucas-Kanade-Tomasi optic flow tracking,
% based on the article:
%
% "Pyramidal implementation of the Lucas Kanade feature tracker -
% description of the algorithm" by Jean-Yves Bouguet.
%
% Goal of the implementation is to understand the algorithm, not to make a
% computationally efficient method for determining the flow.
%
% Steps of the algorithm:
% 1. Determine which points can be tracked between the images
% 2. Build pyramids of the images
% 3. From the top to the bottom level, perform Lucas-Kanade feature
% tracking
% 4. Show the found optic flow
%
% Im1, Im2: two images. If in color, they are transformed to gray scale for
% the optic flow.
% window_size: optional argument for the window size used in the flow
% calculation
% ROI: Optional region in which we detect features in Im1: [xleft, ytop,
% xright, ybottom], default is entire window [1, 1, width, height]
% pyramid_levels: number of pyramid levels, the original image excluded.
% Default number of levels is 0.
% graphics: whether to show the points and vectors in two figures. default
% is true.
% X: possible initialization of feature positions. Default is [].
% V_guess: possible initialization of flow at the feature positions X.
% Default is []. Only works for Bouguet's method, not our own
% implementation.
%
% X: the positions of the features in the first image
% Y: the positions of the features in the second image
% V: corresponding image velocities, corresponding positions in the second
% image are Y = X + V.
% final_errors: errors of the matches after applying flow V
% Q: quality of the points in the first image (only if no X is provided)
%
% (c) Guido de Croon, 07-04-2010

% whether to use BOUGUET's script (gives better results).
% His script can be found here: http://robots.stanford.edu/cs223b04/track_affine.m
BOUGUET = false;

% whether to remove matches that have high errors in terms of pixel value differences:
REMOVE_HIGH_ERRORS = true;

% prepare outputs:
Y = []; V = []; final_errors = []; Q = [];

% get image size:
H = size(Im1,1); W = size(Im1,2);

% process parameters:
if(nargin < 2)
    error('At least 2 images are necessary for running pyramidLKT\n');
end
if(~exist('window_size', 'var') || sum(size(window_size)) < 2)
   window_size = 5; % actually the half_window_size
end
if(~exist('ROI', 'var') || sum(size(ROI)) < 2)
    ROI = [1 1 W H];
end
if(~exist('pyramid_levels', 'var') || sum(size(pyramid_levels)) < 2)
    pyramid_levels = 0;
end
if(~exist('graphics', 'var') || sum(size(graphics)) < 2)
    graphics = true;
end
if(~exist('X', 'var') || sum(size(X)) < 2)
    X = [];
    NO_REMOVAL = false;
else
    NO_REMOVAL = true;
end
if(~exist('V_guess', 'var') || sum(size(V_guess)) < 2)
    V_guess = [];
elseif(isempty(X))
    error('pyramidLKT: If a guessed flow is given, points in the image should also be given.');
end

% transform the images to double-type grayscale images
OIm1 = Im1;
OIm2 = Im2;
Im1 = transformImage(Im1);
Im2 = transformImage(Im2);

% 1. Find features that are fit for tracking:
IM_RESIZE = 1; % image resize for finding the right features
if(isempty(X))
    % resize image:
    if(IM_RESIZE ~= 1)
        Ir = imresize(Im1, IM_RESIZE);
    else
        Ir = Im1;
    end
    % find features in resized image
    [X, Q] = findGoodFeatures(Ir, window_size, ROI);
    % scale back the coordinates:
    X = X * (1 / IM_RESIZE);
end
% number of found points:
n_points = size(X,1);

if(graphics)
    h1 = figure();
    imshow(Im1);
    hold on;
    plot(X(:,1), X(:,2), 'x');
    hold off;
    title('Points in image 1');
end

if(BOUGUET || NO_REMOVAL)
    % 2 & 3 by Bouguet's own script:
    % track_affine(u_list,I,J,Lm,wintx,winty,v_guess,A_guess,affine,normalize,verbiose,DEBUG,resolution,MaxIter);
    if(isempty(V_guess))
        [v_list,A_list,I_patches,J_patches,error_image,a_normalize,b_normalize] = track_affine(X',Im1,Im2, pyramid_levels,[],[],[],[],0);
    else
        [v_list,A_list,I_patches,J_patches,error_image,a_normalize,b_normalize] = track_affine(X',Im1,Im2, pyramid_levels,[],[],X' + V_guess',[],0);
    end
    Y = v_list';
    V = Y - X;
    final_errors = error_image';
else
    % 2. Build pyramids of the images
    Pyramid1 = buildPyramid(Im1, pyramid_levels);
    Pyramid2 = buildPyramid(Im2, pyramid_levels);
    
    if(graphics)
        h4 = figure();
        for pl = 1:pyramid_levels+1
            subplot(2,pyramid_levels+1,pl);
            imshow(Pyramid1{pl});
            title(['level ' num2str(pl-1)]);
            subplot(2,pyramid_levels+1,pyramid_levels+1+pl);
            imshow(Pyramid2{pl});
            title(['level ' num2str(pl-1)]);
        end
    end
    
    % 3. From the top to the bottom level, perform Lucas-Kanade feature tracking
    [Y,V,not_lost, final_errors] = trackFeatures(X, Pyramid1, Pyramid2, window_size);
    if(~NO_REMOVAL)
        X = X(not_lost,:);
        if(REMOVE_HIGH_ERRORS)
            threshold = mean(final_errors);
            low_err = find(final_errors < threshold);
            X = X(low_err,:); Y = Y(low_err,:); V = V(low_err,:);
        end
    end
    n_points = length(X);
end

% 4. show the found optic flow
if(graphics)
    h2 = figure();
    imshow(Im2);
    hold on;
    plot(Y(:,1), Y(:,2), 'x', 'Color', 'red');
    hold off;
    title('Correspondences');

    h3 = figure();
    imshow(0.5 * Im1 + 0.5 * Im2);
    hold on;
    plot(X(:,1), X(:,2), 'x');
    plot(Y(:,1), Y(:,2), 'x', 'Color', 'red');
    if(exist('arrow.m', 'file'))
        % plot arrows for the flow:
        for p = 1:n_points
            if(~isnan(Y(p,1)))
                arrow(X(p,:), Y(p,:), 'Length', 3);
            end
        end
    else
        error('Please download arrow.m - see README.txt.');
    end
%     pause;
%     if(exist('h1', 'var'))
%         close(h1);
%     end
%     if(exist('h2', 'var'))
%         close(h2);
%     end
%     if(exist('h3', 'var'))
%         close(h3);
%     end
%     if(exist('h4', 'var'))
%         close(h4);
%     end
end




