function [flow_data] = test_flow(image_dir, image_type, graphics, step_size, fl_pixels)
% function [flow_data] = test_flow(image_dir, image_type, graphics, step_size, fl_pixels)
%
% Code used in the articles:
% 1. "Optic-flow based slope estimation for autonomous landing",
% by G.C.H.E. de Croon, H.W. Ho, C. De Wagter, E. van Kampen, B. Remes, and Q.P. Chu, 
% International Journal for Micro Air Vehicles, 2013.
%
% 2. "Nonlinear Model Predictive Control Applied to Vision-Based Spacecraft Landing", 
% by D. Izzo and G.C.H.E. de Croon, EuroGNC 2013, Delft, 2013.
%
% for a given directory determines the optic flow and outputs flow data
% (coordinates and flows)
% - image_dir: a string indicating the directory with the images
% - image_type: a string with the image type extension ('jpg' is default)
% - graphics: whether to show the optic flow (false by default)
% - step_size: how far two images that are passed to the optic flow tracker
% - fl_pixels: focal length in pixels
% are apart (default is 1, i.e., subsequent images)
% 
% (c) Guido de Croon, 22-2-2013

if(~exist('image_type', 'var') || isempty(image_type))
    image_type = 'jpg';
end
if(~exist('graphics', 'var') || isempty(graphics))
    graphics = false;
end
if(~exist('step_size', 'var') || isempty(step_size))
    step_size = 1;
end
if(~exist('fl_pixels', 'var') || isempty(fl_pixels))
    n_pixels_camera = 256;
    FOV = 50;
    fl_pixels = 0.5 * n_pixels_camera / tan(0.5 * deg2rad(FOV));
end

% resize factor images, 1 = no resize:
IM_RESIZE = 1;

% whether to show the optic flow tracking graphics:
% LKT_graphics = false;
DEBUG_GRAPHICS = false;

% get the file names of the images:
image_names = makeListOfFiles(image_dir, image_type);
if(exist('sort_nat', 'file'))
    image_names = sort_nat(image_names);
else
    fprintf('Running without sort_nat, file sequence (and hence optic flow determination) may be wrong.\n');
end
n_images = length(image_names);

% create a figure:
if(graphics)
    h(1) = figure();
end

% prepare output flow_data
image_inds = 1:step_size:n_images-step_size;
% flow_data = cell(length(image_inds)-1,1);
j = 1;
for i = image_inds
    
    % print the image number:
    fprintf('Image pair %d, %d (last = %d)\n', i, i+step_size, n_images-step_size);
    
    % load the images:
    Im1 = imread([image_dir '/' image_names{i}]);
    Im2 = imread([image_dir '/' image_names{i+step_size}]);
    Im1 = imresize(Im1, IM_RESIZE);
    Im2 = imresize(Im2, IM_RESIZE);
    
    % perform Lucas-Kanade tracking:
    %     n_pyramid_levels = 2;
    %     [X,Y,V,final_errors] = pyramidLKT(Im1, Im2, [], [], n_pyramid_levels, LKT_graphics);
    %     flow_data{j}.X = X;
    %     flow_data{j}.Y = Y;
    %     flow_data{j}.V = V;
    %     flow_data{j}.final_errors = final_errors;
    
    if(~exist('pu', 'var') || isempty(pu))
        % first time we cannot include estimates of the divergence:
       [FoE, TTC(j), OmegaX(j), OmegaY(j), Zx(j,:), Zy(j,:), pu, pv, FLOW, erroneous(j), inliers(j), errors(j)] = get_TTI_from_flow_Image_Kalman(Im1, Im2, DEBUG_GRAPHICS, [], [], [], []);
    else
        % second time we can:
        [FoE, TTC(j), OmegaX(j), OmegaY(j), Zx(j,:), Zy(j,:), pu, pv, FLOW, erroneous(j), inliers(j), errors(j)] = get_TTI_from_flow_Image_Kalman(Im1, Im2, DEBUG_GRAPHICS, pu, pv, FoE, FLOW);
    end
    
    % show the flow:
    if(graphics)
        X = FLOW.X;
        V = FLOW.V;
        imshow(Im2);
        [H, W] = size(Im2);
        hold on;
        if(exist('arrow.m', 'file'))
            for p = 1:size(X,1)
                % only draw an arrow if it falls within the image:
                if(X(p,1) >= 1 && X(p,2) >= 1 && X(p,1) + V(p,1) >= 1 && X(p,2) + V(p,2) >= 1 && X(p,1) <= H && X(p,2) <= W && X(p,1) + V(p,1) <= H && X(p,2) + V(p,2) <= W)
                    arrow(X(p,:), X(p,:) + V(p,:), 'Length', 5, 'EdgeColor', [1 0 0], 'FaceColor', [1 0 0], 'LineWidth', 2);
                end
            end
        else
            error('Please download arrow.m - see README.txt.');
        end
        hold off;
        drawnow;
        pause(0.5);
    end
    
    j = j + 1;
end

flow_data.TTC = TTC;
flow_data.OmegaX = OmegaX;
flow_data.OmegaY = OmegaY;
flow_data.Zx = Zx;
flow_data.Zy = Zy;
flow_data.erroneous = erroneous;
flow_data.inliers = inliers;
flow_data.errors = errors;

if(graphics)
    % Show the time-to-contact over time:
    h(2) = figure(); 
    set(gcf, 'Color', [1 1 1]);
    plot(TTC, 'LineWidth', 2);
    xlabel('Time (frames)');
    ylabel('Time-to-contact');
    set(h(2), 'Name', 'TTC',  'NumberTitle', 'off');
    
    % Show the slope in the x-direction, slope alpha
    h(3) = figure();
    set(gcf, 'Color', [1 1 1]);
    plotZx = rad2deg(atan(fl_pixels*smooth(Zx(:,1))));
    plot(1:length(Zx), plotZx, 'LineWidth', 2, 'Color', [1 0 0]);
    xlabel('Time (frames)');
    ylabel('Slope \alpha');
    set(h(3), 'Name', 'Slope alpha',  'NumberTitle', 'off');
    
    % Show the slope in the y-direction, slope beta
    h(4) = figure();
    set(gcf, 'Color', [1 1 1]);
    plotZy = rad2deg(atan(fl_pixels*smooth(Zy(:,1))));
    plot(1:length(Zy), plotZy, 'LineWidth', 2, 'Color', [1 0 0]);
    xlabel('Time (frames)');
    ylabel('Slope \beta');
    set(h(4), 'Name', 'Slope beta',  'NumberTitle', 'off');
    
    % Show ventral flow:
    h(5) = figure();
    set(gcf, 'Color', [1 1 1]);
    plot(OmegaX, 'LineWidth', 2, 'Color', [0 0 1]);
    plot(OmegaY, 'LineWidth', 2, 'Color', [0 0 1]);
    xlabel('Time (frames)');
    ylabel('Ventral flow u_0, v_0');
    legend(['u_0', 'v_0']);
    set(h(5), 'Name', 'Ventral flow',  'NumberTitle', 'off');
    
    % Show error over time:
    h(6) = figure();
    set(gcf, 'Color', [1 1 1]);
    plot(errors, 'LineWidth', 2, 'Color', [0 0 1]);
    xlabel('Time (frames)');
    ylabel('Error');
    set(h(6), 'Name', 'Error of fit',  'NumberTitle', 'off');
end
    
