function [Y,V,not_lost,final_errors] = trackFeatures(X, Pyramid1, Pyramid2, window_size)
% function [Y,V,not_lost,final_errors] = trackFeatures(X, Pyramid1, Pyramid2, window_size)
% X is size n x 2
% V is n x 2
% Y is n x 2
% not_lost: indices of points that have not been lost
% 
% (c) Guido de Croon, 2013

graphics = false;

pyramid_levels = size(Pyramid1,1) - 1;

%X = X(1:4:20,:);
n_points = size(X,1);
% initial guess is zero velocity
guessed_flow = zeros(2, n_points);
% initially no points are lost
lost_points = zeros(1, n_points);
epsilon = 1E-5;
% store the final errors, if not lost:
final_errors = zeros(1,n_points);

if(graphics)
   h1 = figure();
   h2 = figure();
   h3 = figure();
end

for pl = pyramid_levels:-1:0
    pl_index = pl + 1;
    WIDTH = size(Pyramid1{pl_index},2); 
    HEIGHT = size(Pyramid1{pl_index},1);
    % determine which points have not yet been lost:
    not_lost = find(lost_points == 0);
    
    % track points one by one:
    for p = not_lost
        % point in the first image, at the current scale and initialised according to the current
        % guess:
        u = [X(p,1); X(p,2)] ./ (2^pl) + guessed_flow(:, p);  % guessed_flow always expressed in the current pyramid level size
        % initialize additional flow for this level:
        v = [0; 0];
        
        % get sub-pixel accurate neighborhood: (what happens if this falls
        % outside of the image?)
        I_padded_neighborhood = getSubPixel(Pyramid1{pl_index}, X(p,1) / (2^pl), X(p,2) / 2^(pl), window_size+1); 
        
        I_neighborhood = I_padded_neighborhood(2:2*window_size+2, 2:2*window_size+2);
        % Determine derivatives of this neighborhood:
        [Ix, Iy] = getGradient(I_padded_neighborhood, 'valid'); 
        
        % TRY-OUT:
        %         J_padded_neighborhood = getSubPixel(Pyramid2{pl_index}, u(1) + v(1), u(2) + v(2), window_size+1);
        %         [Ix, Iy] = getGradient(J_padded_neighborhood, 'valid');
        
        % determine G: (could be done once before iterating, but not if you want to follow points to the border of the image)
        G = getG(Ix, Iy, window_size+1, window_size+1, window_size);
        
        % ***********************
        % Iterative Lucas-Kanade:
        % ***********************
        maxIter = 20;
        if(det(G) < epsilon)
            maxIter = 0;
            lost_points(p) = 1;
        end
        accuracy_threshold = 0.05;
        step = 2 * accuracy_threshold;
        it = 1;
        err = zeros(maxIter,1);
        while(lost_points(p) == 0 && it <= maxIter && sum(abs(step)) > accuracy_threshold && (it <= 5 || err(it-1) < err(it-5))) % this condition may be replaced by a check on the window displacement

            % if the point goes to the border of the image, we have to
            % recalculate G for the valid part of Ix, Iy
            % determine G: (could be done once before iterating, but not if you want to follow points to the border of the image)
            %             G = getG(Ix, Iy, window_size+1, window_size+1, window_size);
            
            % difference in intensity between the two pixels:
            J_neighborhood = getSubPixel(Pyramid2{pl_index}, u(1) + v(1), u(2) + v(2), window_size);
            ImDiff = I_neighborhood - J_neighborhood;
            err(it) = sum(sum(ImDiff.^2));
            % store the error if at the bottom pyramid level
            if(pl == 0)
                final_errors(p) = err(it);
            end
            % fprintf('Error at iteration %d: %f\n', it, err(it));
            % calculate the mismatch vector:
            bx = sum(sum(ImDiff .* Ix));
            by = sum(sum(ImDiff .* Iy));
            b = [bx;by];
            % calculate the optic flow step:
            step = (pinv(G) * b); 
            step = [1;-1] .* step;% in MATLAB, dy negative means up, so we have to convert the y-axis

            if(graphics)
                figure(h1);
                subplot(1,2,1);
                imshow(Pyramid1{pl_index}); hold on; plot(X(p,1) ./ 2^pl, X(p,2) ./ 2^pl, 'x', 'Color', 'red', 'MarkerSize', 5); hold off;
                title(['it = ' num2str(it) ': suggested step = ' num2str(step(1)) ', ' num2str(step(2))]);
                subplot(1,2,2);
                imshow(Pyramid2{pl_index}); hold on; plot(u(1)+v(1), u(2)+v(2), 'x', 'Color', 'red', 'MarkerSize', 5); hold off;
                title(['Error = ' num2str(err(it))]);
                figure(h2);
                subplot(1,2,1); imshow(I_neighborhood); subplot(1,2,2); imshow(J_neighborhood);
                figure(h3);
                imagesc(ImDiff);
                pause(1);
            end
            % fprintf('it = %d, u = (%f,%f), v = (%f, %f)\n', it, u(1), u(2), v(1), v(2));
            
            % update the flow:
            v = v + step;
            
            if(u(1)+v(1) > WIDTH || u(1) + v(1) < 1 || u(2) + v(2) > HEIGHT || u(2) + v(2) < 1)
                % lost point
                lost_points(p) = 1;
                break;
            end
            % increment the iterator:
            it = it + 1;
        end 
        
        % update the guessed flow in the coordinates of this pyramid level:
        guessed_flow(:,p) = guessed_flow(:,p) + v;
        
    end % p = not_lost
    
    
    if(graphics)
        not_lost = find(lost_points == 0);
        figure();
        subplot(1,2,1);
        imshow(Pyramid1{pl_index});
        hold on;
        subplot(1,2,2);
        imshow(Pyramid2{pl_index});
        hold on;

        for p = not_lost
            u = [X(p,1); X(p,2)] ./ (2^pl);
            v = guessed_flow(:, p);
            subplot(1,2,1);
            plot(X(p,1) ./ 2^pl, X(p,2) ./ 2^pl, 'x', 'Color', 'red', 'MarkerSize', 5);
            subplot(1,2,2);
            plot(u(1)+v(1), u(2)+v(2), 'x', 'Color', 'red', 'MarkerSize', 5);

        end
        pause;
    end

    if(pl > 0)
        guessed_flow = 2 * guessed_flow;
    end
    
    
end % pyramid levels

not_lost = find(lost_points == 0);
final_errors = final_errors(not_lost);
V = guessed_flow(:, not_lost)';
Y = X(not_lost,:) + V;